/********************************************************************/
/*                                                                  */
/*   Program Name :  speana                                         */
/*                                                                  */
/*      Version:  2.10  2006-01-31                                  */
/*                                                                  */
/*      Copyright (c) 2002-2006 T.Kondo/NICT All Right Reserved     */
/*                                                                  */
/********************************************************************/
/*
D:\IPVLBI\data\k5vssp32\kas11ft32a.dat 21 1 0 K5/VSSP32 2
  Spectrum analysis of IP-VLBI board data
        (off-line processing)

          T.Kondo 2002.5.20
               Ver 0.00 2002.05.20 
               Ver 1.41 2002.10.01 18000->16000 bug fix
               Ver 1.42 2002.10.11 set default frequency axis linear
               Ver 1.43 2004.10.11 support sampling Frequency > 16MHz
               Ver 1.44 2005.06.12 update due to update in common header name
               Ver 1.45 2005.09.27 Allow to use PGDISP as an environmetal variable
                                   for PPGPLOT device
               Ver 2.00 2005.10.28 Improve processing speed drastically
                                   support K5/VSSP32 data processing
                                   FFTW3.0 is supported if hope so.
               Ver 2.10 2006.01.31 support linear power axis

    How to run:
       speana filename [mode [sekibun [pmode [comment [soffset]]]]]

        where filename -- data file name
                            0 means "tds.data" file
               mode -- YX: axis scale mode
                          X: frequency axis mode
                             =0: log scale  =1: linear scale (default)
                          Y: amplitude axis mode
                             =0: log scale (default)  
                             =1: linear scale (common for 4 channels)
                             =2: linear scale (each max and min)
                      -value: auto correlation (value point center)
                      -1:  auto correlation (full range)
                      
               sekibun  -- integration period (integer second)
                            default is 1 
               pmode -- plot device mode
                       0: XTERM and PostScript faile out (default)
                       1: PostScrip out only 
                       2: XTERM only 
               comment  -- comment (displayed at upper side of graph)
                           (do not include space)
                           If omitted, you will be asked later.
                           NOTE: In the interactive mode, you can sue space.
               soffset -- start offset (sec)
*/
#define SPEANA_VERSION "Ver. 2006-01-31"
#ifdef WIN32
  #define USE_FFTW    // use FFTW package
#endif

#ifdef USE_FFTW
 #ifdef WIN32
  #include "fftw3.h"     // use of FFTW package
 #else
  #include <fftw3.h>     // use of FFTW package
 #endif
#endif

#include "libk5.h"      // common header for IP-VLBI utilities

#include <cpgplot.h>
#include "libfft.h"

/*    Offset from dB to dBm conversion from A/D multibit sampling data */ 
#define OFFSET1  16.0     //1 bit sampling full power convert to 0 dB (theoretical)
      /* following constants are provisional  (measured on 2002/5/18) */
#define OFFSET2   9.0     //2 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET4  -1.0     //4 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET8 -26.0     //8 bit sampling data offset measured at 200kHz for PC(vlbi11)

#define REWIND_BYTE 200 
#define USE_ASM    // use inline assemble function

typedef struct K5VSSP_file     // originated from ip_vlbi_file type
                               // difference is displayed by **K5**
{
    /*
      Note: use of a fpos_t type variable was obsoleted because of
         the compatiblity ploblem of a fpos_t type variable between linux 2.6
         and previous one.
    */
	//unsigned char bit32[4];  // obsoleted in **K5**
	//unsigned char bit64[8]; // obsoleted in **K5**
	unsigned char *ibuf; // will be obsoleted in **K5**
	u_int32_t head32[8]; // **K5**  to contain full header including 
                         // in case of K5/VSSP32
	u_int32_t *aux32; // **K5**  to contain aux field start address in
                        // head 32
	u_int32_t *ubuf; // pointer to read buffer **K5**
    int nword;       // ubuf size allocated or to be allocated  **K5**
    int auxbyte;     // # of bytes in aux files  **K5**
                     // valid byte number in aux32  (deflt=20)
    int headsizew;    // header size in 32bit-word
    int *idata;   // pointer to decoded data (integer) container
    double *data;   // pointer to decoded data (R*8) container
	int sfreq;  //  sampling frequency in kHz
	int adbit;  // AD bits (1,2,4,8)
	int numch;  // # of channels
	int icnt;   //  current unit number in a second data 
	int numb;   //  # of bytes in unit # of samples
	int usampl;  // unit # of samples
	int kpos;  // should be obsoleted ?
	int npos;    //  valid number in idata or xdata **K5**
	int imax;    //  # of units in a second data
	int bytesec;    //  # of bytes in a second data  **K5**
	//char* fname; //pointer data file name obsoleted
	char fname[128]; //data file name opened or to be opened  **K5**
	double fsample;  // sampling frequency (Hz)
	double tsample; // sampling period (sec)
	double unitp;   // usampl*tsampl unit length 
    long int seconds; // time (sec) from mid night (00h00m00s)
	int hh, mm, ss;  // time in hour, minutes, seconds 
    int year; // 4 digits year (K5/VSSP32)
    int totalday; // (K5/VSSP32)
    int version1; // Major version (K5/VSSP32)
    int version2; // Minor version (K5/VSSP32)
	int spatn; // detected 2nd sync pattern detected 
	int syncok; // =1 for 32bit all 1 detection // not used
	int format; // =0 for old format , =1 K5/VSSP, =2 K5/VSSP32
                //         =2 is added **K5**
	FILE *stream;
	//fpos_t pos;   // changed to int64_t type 
	//fpos_t size;  // changed to int64_t type
                    // because fpos_t type is no more scalar value in linux 2.6
	int64_t pos;   // current pointer
	int64_t size;  // opened file size in bytes **K5**
	int fh;
	int ch;    
	int ierr;   // error code (general purpose)
	double dtime;  // time at begin of unit
	double dtimest;  // time at begin of file data
	unsigned int bytesread;  // number of bytes read
	int wordread;   // # of 32bit-words read counter (general) **K5**
	int64_t twordread;   // total # of 32bit-word read **K5**
    int vflag, vflag_old;  // byte slip flag  added on 2003-02-14
                           //  if header sync detection is suceed
                           //     set to 1.  else 0
                           //   1: sync detection success 
                           //   0: sync detection error (may be byte slip)
                           //   vflag_old is the previous status
    long int dtime_vflag, dtime_vflag_old; // time at vflag decision made
} K5VSSP_file_t;

int CheckK5header_with_Read(K5VSSP_file_t *S);

int64_t get_file_size(char *fname)
{
    /*  get file size opened */
    struct stat info;
    stat(fname,&info);
    return (int64_t)info.st_size;
    // use %lld to print out whole number that has int64_t type
}

int header_aux_read(K5VSSP_file_t *K5)
{
   // read Aux field in K5/VSSP32 header
   // assumption : pointer is located just next to 64bit common header

       int i, i32, n;

       i32=0;

       /* next 32bit data read */
       K5->wordread = fread( K5->head32+2, sizeof( u_int32_t ), 1, K5->stream );
       K5->twordread+=K5->wordread;
       K5->pos+=K5->wordread*4;  // pointer position is in bytes
       if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	       i=fclose( K5->stream ) ;
	       i32=-1;
           goto exit_here;
       }
       // parse header into each item
       K5->year=2000+(int)((K5->head32[2] >> 9) & 0x7F); // year
       K5->totalday=(int)(K5->head32[2] & 0x1FF);
       K5->version1=(int)((K5->head32[2] >> 28) & 0xF);
       K5->version2=(int)((K5->head32[2] >> 24) & 0xF);
       K5->auxbyte =(int)((K5->head32[2] >> 16) & 0xFF);
       //printf("year = %d auxbyte = %d\n",K5->year,K5->auxbyte);
       for(n=0; n<K5->auxbyte/4; n++){
           K5->wordread = fread( K5->head32+3+n, sizeof( u_int32_t ), 1, K5->stream );
           K5->twordread+=K5->wordread;
           K5->pos+=K5->wordread*4;  // pointer position is in bytes
           if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	           i=fclose( K5->stream ) ;
		       i32=-1;
               goto exit_here;
	       }
       }
exit_here:
       return i32;

}

// K5 header check (updated from headerchkn2)
int CheckK5header(K5VSSP_file_t *K5)
/*
    K5 header common portion (1st 64bits) check

    if 1st 64bits show K5/VSSP32 pattern,  extended header and
    AUx field are read


    return 0 : K5/VSSP Header detected
          -1 : K5.head32 is not header
         -99 : EOF detected
*/
{
    int sdat[16]={1,2,4,8,0,0,0,0,0,0,0,0,0,0,0,0};
    int fdat[16]={40,100,200,500,1000,2000,4000,8000,16000, 
                  32000,64000,128000,256000,512000,1024000,2048000}; 
    int smplbit0, sfreq0, nch;
    u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
    u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
    int iok;
    iok=-1;

    K5->format=0;  // suppose K5/VSSP old format
    if(K5->head32[0]==0xFFFFFFFFUL){
        if((K5->head32[1] & 0xFF000000UL) == syncpat) K5->format=1; // this is sync block
        if((K5->head32[1] & 0xFF000000UL) == syncpat32) K5->format=2; // this is sync block
        //if((K5->head32[0]==0xFFFFFFFFUL) && ((K5->head32[1] & 0xFF000000UL) == syncpat)){ // this is sync block
        //if((K5->head32[0]==0xFFFFFFFFUL) && ((K5->head32[1] & 0xFF000000UL) == syncpat32)){ // this is sync block
        if(K5->format > 0){ 
            iok=0;
            /* parse sampling information and get time data */ 
            K5->spatn=(int)((K5->head32[1] >> 24) & 0xFF);       
            smplbit0 = (K5->head32[1] >> 22) & 0x3;   
            K5->adbit = sdat[smplbit0];      // AD bits
            sfreq0 = ( K5->head32[1] >> 18) & 0xF;
            K5->sfreq=fdat[sfreq0];
            nch = (K5->head32[1] >> 17) & 1;
            if (nch == 1) {
                K5->numch = 4; 
            } else {
                K5->numch = 1;
            }
            K5->seconds=(long int)(K5->head32[1] & 0x1FFFFUL);
            sec2hms(K5->seconds,&K5->hh,&K5->mm,&K5->ss);
        }
        if(K5->format==2){  // read further header
            K5->headsizew=8;
            if(header_aux_read(K5)<0) iok=-99;
       }

    }
    return iok;
}

#ifdef USE_ASM

    #ifdef WIN32
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)  // for VC++ inline assembler
        {
        /*
                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        
                ebp+14h --  pat

        */
                __asm{
                    mov edx, ilast
                    mov eax, K5pos
                    mov ebx, wd
                    mov cl, byte ptr [adbit]
                    //mov ch, byte ptr [pat]  //not necessary
                    xor esi,esi  
            loop_1:
                    mov edi, ebx
                    and edi, pat
                    mov [eax+esi*4], edi
                    shr ebx, cl
                    inc esi
                    cmp esi, edx
                    jnz loop_1
                }
        }
    #else  // for gcc
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
        {
        /*
                K5pos --  8(%ebp)
                wd    --  12(%ebp)
                ilast --  16(%ebp)
                pat   --  20(%ebp)
                adbit --  24(%ebp)

                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        

        */
            asm volatile (
                 "pushl %eax \n\t"     // save register used in this function
                 "pushl %ebx \n\t"
                 "pushl %ecx \n\t"
                 "pushl %edx \n\t"
                 "pushl %edi \n\t"
                 "pushl %esi \n\t"

                 "movl 16(%ebp), %edx \n\t"
                 "movl  8(%ebp), %eax \n\t"
                 "movl 12(%ebp), %ebx \n\t"

                 "movb 24(%ebp), %cl \n\t"
                 "xor %esi, %esi \n\t"  
            "loop_10: \n\t"
                 "movl %ebx, %edi \n\t"
                 "and  20(%ebp), %edi \n\t"
                 "movl %edi, (%eax,%esi,4) \n\t"
                 "shr %cl, %ebx  \n\t"
                 "inc %esi \n\t"
                 "cmp %edx, %esi \n\t"
                 "jnz loop_10 \n\t"
                 "popl %esi \n\t"    // recover saved register 
                 "popl %edi \n\t"
                 "popl %edx \n\t"
                 "popl %ecx \n\t"
                 "popl %ebx \n\t"
                 "popl %eax \n\t"
                );
        }
    #endif
#else  // normal C function
    void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
    {
        int i,ww, wd1;
        wd1=wd;
        for(i=0; i<ilast; i++){
            ww=(int)(wd1 & pat);
            //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
            K5pos[i]=ww;
            wd1>>= adbit ; 
        }
    }
#endif

void parseADdata_3N(K5VSSP_file_t *K5)
{
/*
    decode K5/VSSP (IP-VLBI) data into integer data

      [in]
         K5->ubuf    ---  sampler (K5/VSSP (IP-VLBI) board) out data 
         K5->wordread --  # of data in K5->ubuf (32bit word unit)
         K5->adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         K5->idata -- data parsed into each sampled data
         K5->npos  -- valid number of data in xdata array

     Relation between # of ch (numch) and each ch time siries

        sdata(m,k)=xdata(k*numch+m)

             where  numch :  # of channels  (1, 4)
                      m   :  CH#
                      k   :  k-th sampling
        
*/
     int numb8, pat, n, ilast, ipos;
     int pattbl[9]={1,1,3,3,15,15,15,15,255};
     u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
     u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
     u_int32_t wd1,wd2;
     int *K5pos;
     unsigned char bpat;

     if(K5->vflag != 1) goto exitend;
     //numb8=numb;
     numb8=K5->wordread*4;

     ipos=0;
     ilast=32/K5->adbit;
     pat=pattbl[K5->adbit];
     bpat=(unsigned char)pat;
     wd1=K5->ubuf[0];

     //printf("Debug IN PARSE adbit, pat %d %d\n",K5->adbit,pat);
     K5pos=K5->idata;
     for( n=1; n < K5->wordread; n++){
        wd2=K5->ubuf[n];
        /* check sync pattern */
        if ( (wd1 == 0xFFFFFFFFUL) && (((wd2 >> 24) & 0xFF) == (K5->spatn & 0xFFUL)) ) { // sync detected for K5/VSSP
            printf("Sync detected \n"); 
            continue;  // pass decoding
        }
//        for(i=0; i<ilast; i++){
//             ww=(int)(wd1 & pat);
//             //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
//             K5->idata[ipos]=ww;
//             wd1=wd1 >> K5->adbit ; 
//             ipos++;
//        }
        /* use of inline assembler */
        K5pos=&K5->idata[ipos];
        decode_data(K5pos, wd1, ilast, pat, K5->adbit);   // use of inline assembler
        //decode_data2(K5pos, wd1, ilast, pat, K5->adbit); // normal C function
                              // no big difference found between decode_data and decode data2
                              //   ex. decode_data  34.279 sec for 8MHz 1bit 4CH 20 sec data
                              //   ex. decode_data2 34.289 sec for 8MHz 1bit 4CH 20 sec data
                              //  direct use like below is also almost same processing speed
                              //     i.e. 34.289 sec
/*
        __asm{
            mov edx, ilast
            mov eax, K5pos
            mov ebx, wd1
            mov cl, bpat
            xor esi,esi  
    loop_1:
            mov edi, ebx
            and edi, pat
            mov [eax+esi*4], edi
            shr ebx, cl
            inc esi
            cmp esi, edx
            jnz loop_1
        }
*/
        ipos+=ilast;

        wd1=wd2;
        //K5pos+=ilast;

     }
/*     
     for(i=0; i<ilast; i++){
         ww=(int)(wd1 & pat);
         //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
         K5->idata[ipos]=ww;
         wd1=wd1 >> K5->adbit ; 
         ipos++;
     }
*/     

     //K5pos=K5->idata+ipos;
     K5pos=&K5->idata[ipos];
     decode_data(K5pos, wd1, ilast, pat, K5->adbit);  
     //decode_data2(K5pos, wd1, ilast, pat, K5->adbit);
     /* use of inline assembler */
/*
     __asm{
        mov edx, ilast
        mov eax, K5pos
        mov ebx, ww
        mov cl, bpat
        xor esi,esi  
loop_2:
        mov edi, ebx
        and edi, pat
        mov [eax+esi*4], edi
        shr ebx, cl
        inc esi
        cmp esi, edx
        jnz loop_2
     }
*/
     ipos+=ilast;

     K5->npos=ipos;

exitend:;
     //printf("numb8,ipos %d %d\n",numb8,ipos);
}

int CheckK5header_with_Read(K5VSSP_file_t *S)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of words read till header finding
               -1    no header found till EOF
    */
      
       int i, i32;
       i32=0;

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32

       /* first 32bit read */

       S->wordread = fread( S->head32, sizeof( u_int32_t ), 1, S->stream );
       S->twordread+=S->wordread;
       S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //printf("%X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* next 32bit data read */
       S->wordread = fread( S->head32+1, sizeof( u_int32_t ), 1, S->stream );
       S->twordread+=S->wordread;
       S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
       } else {
           S->head32[0]=S->head32[1];
           i32++;
       }
       goto loop1;
exit_loop:
       if(i32 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i32;
}

int CheckK5header_with_Read2(K5VSSP_file_t *S, int maxword)
{
    /* 
       Read 64bit Data header check to get sampling parameters 
    
       if first 64bit data is not header block, read file at maximum 
       maxword 32bit-words untill finding header block.

       Return
               +ve   # of words  read till header finding
               -1    no header found after maxword read
    */
      
       int i, i32;
       i32=0;

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       /* first 32bit read */

       S->wordread = fread( S->head32, sizeof( u_int32_t ), 1, S->stream );
       S->twordread+=S->wordread;
       S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //printf("%X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* next 32bit data read */
       S->wordread = fread( S->head32+1, sizeof( u_int32_t ), 1, S->stream );
       S->twordread+=S->wordread;
       S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //headerchkn(S->bit64,&iok,&S->spatn,&S->adbit,&S->sfreq,&S->numch,
       //                 &S->seconds,&S->hh,&S->mm,&S->ss);
       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
       } else {
           S->head32[0]=S->head32[1];
           i32++;
           if(i32+2 > maxword) return -1;
       }
       goto loop1;
exit_loop:
       if(i32 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i32;
}


int fwdNsecK5(int nsec, K5VSSP_file_t *K5)
{
    /* 
       forward fast N sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind 1sec data
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one

    */
    int i,k;
    int maxword;
    int nbytes, nback;
    int64_t tword;
	int64_t pos;

    if(nsec < 1) return 0; // nothing to do

    tword=K5->twordread;  // save current word read counter data 
    pos=K5->pos;  // save current position

    k=0;

    nbytes=K5->bytesec*nsec;
    K5->pos+=nbytes;  // pointer position to be set
    if(K5->pos < K5->size){  // not yet reached to EOF
         fseek(K5->stream,nbytes,SEEK_CUR); // move pointer by K5->bytesec
         if( ferror( K5->stream ) ){
            return -1;
         }
    } else {  // exceed EOF so pinter set to EOF
         fseek(K5->stream,(int)(K5->size - pos),SEEK_CUR); // 
         if( ferror( K5->stream ) ){
            return -1;
         }
        K5->wordread=(int)((K5->size - pos)/4); // convert to 32bit-word unit
        K5->twordread+=(int64_t)K5->wordread;
        K5->pos+=K5->wordread*4;  // pointer position is in bytes
        return -1;
    }
    K5->wordread=nbytes/4;
    K5->twordread+=(int64_t)K5->wordread;


    /* read 64 bit header */
    K5->wordread = fread(K5->head32, sizeof(u_int32_t), 2, K5->stream);
    K5->twordread+=K5->wordread;
    K5->pos+=K5->wordread*4;  // pointer position is in bytes
    if( ferror( K5->stream ) || K5->wordread==0 ){ // EOF or File read error
	   return -1;
    }
    if((i=CheckK5header(&(*K5))) < 0){  // this is not the case of header
       if(i==-99) return -1;
       /* rewind file */
       if(REWIND_BYTE == 0){
           fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
           K5->twordread=tword;            // word counter retrieve
           K5->pos=pos;  // retrieve initial position
           if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
               return -1;
           }
       } else {
           maxword=REWIND_BYTE/4;
           fseek(K5->stream,-(int)REWIND_BYTE,SEEK_CUR); // set position at initial position
           K5->twordread-=REWIND_BYTE/4;            // word counter retrieve
           K5->pos-=REWIND_BYTE;  // retrieve initial position
           /* 1st trial */
           if((k=CheckK5header_with_Read2(&(*K5),maxword)) < 0){ // EOF or file read error
               /* 2nd trial */
               if(nsec==1){
                   nback=(int)(K5->pos-pos);  // rewind to initial position
               } else {
                   nback=K5->bytesec;   // rewind just 1 sec
               }

               fseek(K5->stream,-nback,SEEK_CUR); // set position at initial position
               K5->twordread-=nback/4;            // word counter rewind
               K5->pos-=nback;  // initial position rewind
               if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
                   return -1;
               }
           }
       }
       // get headder after k words read
    }
       
    return k;
}

int fwd1secK5(K5VSSP_file_t *K5)
{
    /* 
       forward fast 1 sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind previous header position
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one
    */
    int i,k;
    int maxword;
    int64_t tword;
	int64_t pos;
    tword=K5->twordread;  // save current word read counter data 
    pos=K5->pos;  // save current position

    k=0;

    K5->pos+=K5->bytesec;  // pointer position to be set
    if(K5->pos < K5->size){  // not yet reached to EOF
         fseek(K5->stream,K5->bytesec,SEEK_CUR); // move pointer by K5->bytesec
         if( ferror( K5->stream ) ){
            return -1;
         }
    } else {  // exceed EOF so pinter set to EOF
         fseek(K5->stream,(int)(K5->size - pos),SEEK_CUR); // 
         if( ferror( K5->stream ) ){
            return -1;
         }
        K5->wordread=(int)((K5->size - pos)/4); // convert to 32bit-word unit
        K5->twordread+=(int64_t)K5->wordread;
        K5->pos+=K5->wordread*4;  // pointer position is in bytes
        return -1;
    }
    K5->wordread=K5->bytesec/4;
    K5->twordread+=(int64_t)K5->wordread;

    /* read 64 bit header */
    K5->wordread = fread(K5->head32, sizeof(u_int32_t), 2, K5->stream);
    K5->twordread+=K5->wordread;
    K5->pos+=K5->wordread*4;  // pointer position is in bytes
    if( ferror( K5->stream ) || K5->wordread==0 ){ // EOF or File read error
	   return -1;
    }
    if((i=CheckK5header(&(*K5))) < 0){  // this is not the case of header
       if(i==-99) return -1;
       /* rewind file */
       if(REWIND_BYTE == 0){
           fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
           K5->twordread=tword;            // word counter retrieve
           K5->pos=pos;  // retrieve initial position
           if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
               return -1;
           }
       } else {
           maxword=REWIND_BYTE/4;
           fseek(K5->stream,-(int)REWIND_BYTE,SEEK_CUR); // set position at initial position
           K5->twordread-=REWIND_BYTE/4;            // word counter retrieve
           K5->pos-=REWIND_BYTE;  // retrieve initial position
           /* 1st trial */
           if((k=CheckK5header_with_Read2(&(*K5),maxword)) < 0){ // EOF or file read error
               /* 2nd trial */
               fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
               K5->twordread=tword;            // word counter retrieve
               K5->pos=pos;  // retrieve initial position
               if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
                   return -1;
               }
           }
       }
       // get headder after k words read
    }
       
    return k;
}

void plotspe_a(float *spow, int npos, int sfreq, int adbit, 
               char *lab, char *lab2, int mode2,
               float ymaxdb, float ymindb)
/*
               mode2 -- YX: axis scale mode
                          X: frequency axis mode
                             =0: log scale  =1: linear scale
                          Y: amplitude axis mode
                             =0: log scale 
                             =1: linear scale (common for 4 channels)
                                  (use ymaxdb, ymindb)
                             =2: linear scale (each max and min)
*/
{
      float fmax, fmin, fsampl, df, ymax, ymin, pmax, pmin;
      float fscale;
      float *fdat, *powdat;
      char htitl[60], xtit[60], ytit[60];
      int n, npmax;
      int mode,ymode;

      ymode=mode2/10;
      mode=mode2 % 10;
      strcpy(htitl,"POWER SPECTRUM");

      /* frequency axis generation */

      npmax=npos/2;

      fdat   = (float *)malloc(npmax*sizeof(float));
      powdat = (float *)malloc(npmax*sizeof(float));

      /* frequency axis */
      fsampl=(float)1000.0*(float)sfreq;
      df=fsampl/(float)npos;
      if(sfreq > 8000){  // if sampling frequency exceeds 8 MHz
          fscale = 1.0e6;  // scaling in MHz
          strcpy(xtit,"FREQUENCY (MHz)");
      } else {
          fscale = 1000.0; // scaling in kHz
          strcpy(xtit,"FREQUENCY (kHz)");
      }
      if(mode == 0) {    // log scale
         fmax=(float)log10(fsampl/(2.0*fscale));
         fmin=(float)log10(df/fscale);
         for(n=0; n<npmax; n++){fdat[n]=(float)log10((float)n*df/fscale); }
      } else {          //  linear scale
         fmax=(float)(fsampl/(2.0*fscale));
         fmin=0.0;
         for(n=0; n<npmax; n++){fdat[n]=(float)((float)n*df/fscale); }
      }

      /* amplitude axis */

      /* data copy, if necessary convert to linear scale */
      if(ymode == 0){    // log scale
         for(n=0; n<npmax; n++) powdat[n]=spow[n];
      } else {   // linear scale
         /* convert to linear scale */
         for(n=0; n<npmax; n++) powdat[n]=(float)pow(10.0,spow[n]/10.0);
      }

      /* data max, min check */
      for(n=0; n<npmax; n++){
         //powdat[n]=spow[n];
         if(n==0){
            pmax=powdat[n];
            pmin=powdat[n];
         } else {
            if(powdat[n] > pmax){ pmax=powdat[n]; }
            if(powdat[n] < pmin){ pmin=powdat[n]; }
         }
      }

      if(ymode == 0){    // log scale
         strcpy(ytit,"POWER (dBm)");
         if( adbit==1 ){strcpy(ytit,"POWER (dB)"); }
         ymax=10.0;
         ymin=-80.0;
         if (adbit == 8 ) ymin=-100;     // extend dynamic range 
      } else {   // linear scale
         strcpy(ytit,"POWER (linear)");
         ymax=pmax;
         ymin=pmin;
         if(ymode==1){
            ymax=(float)pow(10.0,ymaxdb/10.0);
            ymin=(float)pow(10.0,ymindb/10.0);
         }
      }


      //maxdbm=pmax;
      //mindbm=pmin;

      //goto exitend; //debug

      /* open graphics device */
      /*   move to main */
       //if (cpgopen("?") < 1)
      // if (cpgopen("/GF") < 1)
      // {
      //    goto exitend;
      // }

      cpgsch(2.0);  // set text height 2.0  (default is 1)
      cpgslw(2);    // set line width

      /* define coordinate range of graph */
      if(mode == 0){   
         cpgenv(fmin, fmax, ymin, ymax, 0, 10);   // log scale
      } else {         
         cpgenv(fmin, fmax, ymin, ymax, 0, 0);   // linear scale
      }
      
      /* label the axes */
       cpglab(xtit,ytit,htitl);

      /* sub label write*/
       cpgsch(1.5);
       cpgmtxt("T",1.5,0.5,0.5,lab);
       cpgmtxt("T",0.5,0.5,0.5,lab2);
       cpgsch(2.0);
      
      /* check data */
      if (pmax == pmin){
          cpgmtxt("B",-6.0,0.5,0.5,"NO TIME VARIABLE");
          cpgmtxt("B",-4.0,0.5,0.5,"COMPONENT FOUND");
          /* label the axes */
          
      } else {

          /* plot the line graph */
          fdat[0]=fdat[1];
          powdat[0]=powdat[1];  // to avoid ugry plot
          //cpgslw(1);    // set line width
          cpgline(npmax,fdat,powdat); 

      }
      /* close the graphics device */
      /* move to main */
      //cpgclos();
  //exitend: 
     free(powdat);
     free(fdat);
}

void plotautocor(float *spow, int npos, int sfreq, int adbit, char *lab, char *lab2, int amode)
/*
        mode = 0; log frequency scale   1; linear frequency scale
*/
{
      //float cofst,tmax, tmin, fsampl, dt, ymax, ymin, pmax, pmin, maxdbm, mindbm;
      float tmax, tmin, fsampl, dt, ymax, ymin, pmax, pmin;
      float *tdat, *pdat;
      //float x,y;
      char htitl[60], xtit[60], ytit[60];
      int n, k, npmax, nhalf;

      strcpy(htitl,"AUTO CORRELATION (ABS)");

      /* time axis generation */

      if(amode <= 1) {
         npmax=npos;
      } else {
         npmax=amode;
      }
      tdat = (float *)malloc(npmax*sizeof(float));
      pdat = (float *)malloc(npmax*sizeof(float));
      nhalf=npmax/2;

      fsampl=(float)(1000.0*(float)sfreq);
      dt=(float)(1.0/fsampl);
      for(n=0; n<npmax; n++){
          tdat[n]=(float)((float)(n-nhalf)*dt*1.0e6);   // usec unit
      }
      tmax=tdat[npmax-1];
      tmin=tdat[0];

      strcpy(xtit,"DELAY TIME (us)");
      strcpy(ytit,"CORRELATION AMP");
      ymax=(float)1.1;
      ymin=0.0;

      /* data range limited copy and max, min check */
     
      for(n=0; n<npmax; n++)
      {
         k=n-nhalf;
         if(k < 0) k=npos+k;
         pdat[n]=spow[k];
         if(pdat[n]<0) pdat[n]=-pdat[n];  // to take absolute value
         if(n==0)
         {
            pmax=pdat[n];
            pmin=pdat[n];
         }
         else
         {
            if(pdat[n] > pmax){ pmax=pdat[n]; }
            if(pdat[n] < pmin){ pmin=pdat[n]; }
         }
      }
      

      cpgsch(2.0);  // set text height 2.0  (default is 1)
      cpgslw(2);    // set line width

      /* define coordinate range of graph */
      cpgenv(tmin, tmax, ymin, ymax, 0, 0);   // linear scale
      
      /* label the axes */
       cpglab(xtit,ytit,htitl);

      /* sub label write*/
       cpgsch(1.5);
       cpgmtxt("T",1.5,0.5,0.5,lab);
       cpgmtxt("T",0.5,0.5,0.5,lab2);
       cpgsch(2.0);
      /* check data */
      if (pmax == pmin)
      {
          cpgmtxt("B",-6.0,0.5,0.5,"NO DATA FOR PLOT");
          
      }
      else
      {
      
        /* plot the line graph */
         cpgline(npmax,tdat,pdat); 
      }

      /* move to main */
  //exitend: 
     free(pdat);
     free(tdat);
}


int main(int argc, char *argv[])
{
    int sfreq, adbit, numch;
    long int startoffset; // start time offset in sec
    int nloop,sekibun;
    int i,imax, m, k, kaisu;
    int smode,usampl,nfft,kplot;
    int kword;
    int bits, bytes, numb;
    int amode;    // >0 auto correlation plot mode 
    int mode;  // frequency axis mode  0: log  1: linear 
    int pmode;
    int mode2;  // 2 digits axis mode
    int ymode; // amplitude scale mode   0: log   1: linear
    int scount[4];
    float fsampl,tsampl,fact;
    double *cdata;
    double *f_hanning;
    double theta;
    float *xdata, *ydata;
    double sum,dc;
    float pows, *spows;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;
#ifdef USE_FFTW
    fftw_complex *cc;
    fftw_plan pfcc, pbcc;
#endif
    double nfact;
    double telapsed;           // to measure processing time
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char fname[60];
    char lab[60], lab2[60], iplab[40], ltime1st[60];
    char labfsampl[10];
    char pgdevice[20];
    char* dflt_fname="tds.data";
    K5VSSP_file_t K5;

    /* set graphic device */
    strcpy(pgdevice,select_env_dflt(PGDISP, PGDISP_DFLT));

    if(argc == 1){
          printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#else
          printf("\n");
#endif
          printf("Usage: speana filename [mode [sekibun [pmode [comment [soffset]]]]]\n");
          printf("\n");
          printf("         where  filename ---- Data file name\n");
          printf("                              if 0 is given, tds.data will be used\n");
          printf("                mode -------- set axis mode or autocorrelation mode\n");
          printf("                              =0 : amplitude log, frequency log\n");
          printf("                              =1 (deflt) : amplitude log, frequency linear\n");
          printf("                              =10 : amplitude linear, frequency log\n");
          printf("                              =11 : amplitude linear, frequency linear\n");
          printf("                              =-1 : Auto correlation function plot\n");
          printf("                                     instead of spectrum\n");
          printf("                              =-N : N points auto correlation function plot\n");
          printf("                sekibun --- Averaging period (integer seconds)\n");
          printf("                             default is 1\n");
          printf("                pmode ---- Plot device selection\n");
          printf("                           =0 :  %s and PostScript file (pgplot.ps) out\n",pgdevice);
          printf("                           =1 :  PostScript file (pgplot.ps) out only\n");
          printf("                           =2 :  %s only\n",pgdevice);
          printf("                          default is 0\n");
          printf("                comment --- Comment appeared on a graph\n");
          printf("                            if omitted, you will be asked later.\n");
          printf("                soffset --- Start time offset in sec (integer seconds)\n");
          printf("                             default is 0\n"); 
          printf("\n");
          printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
          printf("                   = %s",getenv(PGDISP));
          if(getenv(PGDISP)==NULL) {
              printf(",  program deflt is ( %s )\n",PGDISP_DFLT);
          } else {
              printf("\n");
          }
          printf("\n");
          exit(0);
    } 

    strcpy(fname,*(argv+1));
    if (fname[0]=='0'){
          strcpy(fname,dflt_fname);
    }
    strcpy(K5.fname,fname);
    /* plot mode default set */
    mode2=1;
    mode=1;   // default frequency axis mode (=Linear sacle)
    ymode=0;  // default amplitude axis mode (=LOG scale)
    amode=0;  // auto correlation mode off
    /* plot mode set from operator */
    if (argc>=3){
       mode2 = atoi(*(argv + 2 ));
       if(mode2 <0 ){
            amode=-mode2;
            mode=0;
       } else {
            amode=0;
            ymode=mode2/10;
            mode=mode2 % 10;
            
       }
    }

    if (argc>=4){
       sekibun = atoi(*(argv + 3 ));
    } else {
          sekibun=1;     // default integration time (sec)
    }

    if (argc>=5){
       pmode = atoi(*(argv + 4 ));
    } else {
          pmode=0;     // default plot mode
    }

    printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
    printf(" compliled with FFTW3.0 \n");
#else
    printf("\n");
#endif

    cstart1sec = clock();

    printf("Data File is %s\n",fname);
    K5.size=get_file_size(K5.fname);
    if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ){
         //perror( "̓t@CJ܂" );
         //exit( 1 );
      // return -1;
         printf("File (%s) Open Error! \n",K5.fname);
        goto exitend2;
    }
    K5.twordread=0;  // reset word read counter
    K5.pos=0;        // reset current pointer position (in bytes)

    /* get comment */

    if (argc>=6){
          strcpy(iplab,*(argv+5));
    } else {
          /* enter board ID */
          /*
          printf("Enter IP-VLBI BOARD ID -> ");
          read1line(iplab);      // 1 line read from stdin
          printf("Board ID is %s\n",iplab);      
          strcpy(lab2,"IP-VLBI Board ID : ");
          strcat(lab2,iplab);
          */

          /* enter any comment */

          printf("Enter any comment -> ");
          read1line(iplab);      // 1 line read from stdin
          //printf("Board ID is %s\n",iplab);      
          //strcat(lab2,iplab);
    }
    if (argc>=7){
       startoffset = (long int)atoi(*(argv + 6 ));
    } else {
       startoffset=0;     // default startoffset
    }

    /* Data header check to get sampling parameters */
    if((kword=CheckK5header_with_Read(&K5)) < 0){
        printf("File read error! \n");
        //fgetpos(K5.stream,&K5.pos); 
        i=fclose( K5.stream ) ;
        goto exitend;
    }

//nextstep:
    /* set graphic device */
    strcpy(pgdevice,select_env_dflt(PGDISP, PGDISP_DFLT));

    /* parameter set */
    sfreq=K5.sfreq;
    adbit=K5.adbit;
    numch=K5.numch;
    fsampl=(float)(1000.0*(float)sfreq);   // sampling frequency (Hz)
    tsampl=(float)(1.0/fsampl);            // sampling period (sec)
    bits=((long)fsampl)*adbit*numch;  // # of bits in 1 sec data except header
    bytes=bits/8;                     // # of bytes in 1 sec data except header
    K5.fsample=fsampl;
    //sekibun=1;                       // Averaging period for Spectrum (sec)
    K5.bytesec=bits/8;                  //# of bytes in 1 sec data

    smode=1;                         // low resolution mode

    if(smode ==1){ 
        usampl=10000; 
    } else {
        usampl=50000;
        if(sfreq == 40)  usampl=40000; 
        if(sfreq == 100) usampl=100000; 
        if(sfreq == 200) usampl=200000; 
    }
    imax=(sfreq*1000)/usampl;
    numb=usampl*adbit*numch/8;   // # of bytes in unit # of samples
    K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
    K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second

    printf("imax numb  %d %d\n",K5.imax,K5.numb);    //debug

    nfft=1024;
    if(usampl < 524288) nfft= 524288;
    if(usampl < 262144) nfft= 262144;
    if(usampl < 131072) nfft= 131072;
    if(usampl < 65536)  nfft= 65536;
    if(usampl < 32768)  nfft= 32768;
    if(usampl < 16384)  nfft= 16384;

    nfact=2.0/(double)nfft;

    K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl
    K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

    K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
    K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

#ifdef USE_FFTW
    cc=fftw_malloc(sizeof(fftw_complex)*nfft);

    pfcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_FORWARD, FFTW_ESTIMATE);
    pbcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

    //ibuf = (unsigned char *)malloc(numb);
    //idata = (int *)malloc(4*usampl*sizeof(int));
    cdata = (double *)malloc(2*nfft*sizeof(double));
    spows = (float *)malloc(4*nfft*sizeof(float));
    xdata = (float *)malloc(nfft*sizeof(float));
    ydata = (float *)malloc(nfft*sizeof(float));
    f_hanning = (double *)malloc(nfft*sizeof(double));
    //autocor=(float *)malloc(4*nfft*sizeof(float));

    //Hanning filter coeeficients prepare
    for(k=0; k<nfft; k++){
        theta=360.0*(double)k/(double)(nfft-1) - 180.0;
        theta=fmod(theta,360.0) ;
        f_hanning[k]=0.5*(1.0+cos(theta*PIRA));
    }

    fact=((float)usampl)*((float)usampl);
        
    /* sampling frequency in text for print */
    if(sfreq < 1000){
       sprintf(labfsampl,"%dkHz\0",sfreq);
    } else {
       sprintf(labfsampl,"%dMHz\0",sfreq/1000);
    }


    /* foward startffset seconds */
    if(fwdNsecK5(startoffset, &K5) <0){
        printf("Foward error %d\n",startoffset);
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    for(k=0; k<4*nfft; k++) spows[k]=0.0;
    for(k=0; k<4; k++) scount[k]=0;

    printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
        K5.hh,K5.mm,K5.ss,K5.seconds);
    sprintf(ltime1st,"Time(UT) %02d:%02d:%02d\0",K5.hh,K5.mm,K5.ss);
    strcpy(lab2,ltime1st);
    strcat(lab2,"  ");
    strcat(lab2,iplab);
        

    nloop=0;
loop1:
    if(nloop >0) {
        if((kword=CheckK5header_with_Read(&K5)) < 0){
            i=fclose( K5.stream ) ;
            goto exitend;
        }
        printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
            K5.hh,K5.mm,K5.ss,K5.seconds);
    }

    for(i=0; i<K5.imax; i++){
        K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
        K5.pos+=K5.wordread*4;
        K5.vflag=1; // validiate data
        parseADdata_3N(&K5);
        for(m=0; m<numch; m++){
            kaisu=0;
            sum=0.0;

            // initialize array
#ifdef USE_FFTW
            for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
#else
            for(k=0; k<2*nfft; k++) cdata[k]=0.0;
#endif

            for(k=0; k<K5.npos/K5.numch; k++){

                /* set data */
#ifdef USE_FFTW
                cc[k][0]=(double)K5.idata[k*numch+m];  
                sum+=cc[k][0];
#else
                cdata[2*k]=(double)K5.idata[k*numch+m];  
                sum+=cdata[2*k];
#endif
                kaisu++;
            }
             
            dc=sum/(double)kaisu;    // DC component
     
            // DC offset removal from data
            for(k=0; k<K5.npos/numch; k++){
#ifdef USE_FFTW
                cc[k][0]-=dc;
#else
                cdata[2*k]-=dc;
#endif
            }
            //hanning(cdata,nfft);  // hanning filter
            //cdft(2*nfft,-1,cdata);
            if(amode ==0) {    // spectrum mode
                for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                   cc[k][0]*=f_hanning[k];  // hanning filter
                   cc[k][1]*=f_hanning[k];  // hanning filter
#else
                   cdata[2*k]*=f_hanning[k];
                   cdata[2*k+1]*=f_hanning[k];
#endif
                }
                //hanning(cdata,usampl);  // hanning filter
#ifdef USE_FFTW
                fftw_execute(pbcc);
#else
                cdft(2*nfft,1,cdata);
#endif
            } else {
#ifdef USE_FFTW
                fftw_execute(pfcc);
                for(k=0; k<nfft; k++){
                    cc[k][0]*=nfact; cc[k][1]*=nfact;
                }
#else
                cdft(2*nfft,-1,cdata);
#endif
            }
            for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                pows=(float)((cc[k][0]*cc[k][0]+cc[k][1]*cc[k][1])/(double)fact);
#else
                pows=(float)((cdata[2*k]*cdata[2*k]+cdata[2*k+1]*cdata[2*k+1])/(double)fact);
#endif
                spows[k*4+m]=spows[k*4+m]+pows;
            }

            scount[m]++;
        }
    }
    nloop++;
    if(nloop < sekibun) goto loop1;
    i=fclose( K5.stream ) ;
    //printf("%d bytes read\n",pos); 

    if(amode > 0) {   // auto correlation function calc
       for(m=0; m<numch; m++){
          for(k=0; k<nfft; k++){
             spows[k*4+m]/=scount[m]; // nomalization
#ifdef USE_FFTW
             cc[k][0]=spows[k*4+m]; cc[k][1]=0.0;
#else
             cdata[2*k]=spows[k*4+m];
             cdata[2*k+1]=0.0;
#endif
          }
          if(m==1 || m==3){
#ifdef USE_FFTW
              cc[0][0]=1.0;
#else
              cdata[0]=1.0;  // to avoid math error
#endif
          }
#ifdef USE_FFTW
          fftw_execute(pbcc);
#else
          cdft(2*nfft,1,cdata);
#endif
          for(k=0; k<nfft; k++){
             // if(k<10) printf("%d %e\n",k,cdata[2*k]);  // debug
             // if(k>nfft-10) printf("%d %e\n",k,cdata[2*k]);  // debug
#ifdef USE_FFTW
             if(k==0) pmax=(float)cc[0][0];
             if(cc[k][0] > pmax) pmax=(float)cc[k][0];
             spows[k*4+m]=(float)cc[k][0];
#else
             if(k==0) pmax=(float)cdata[0];
             if(cdata[2*k] > pmax) pmax=(float)cdata[2*k];
             spows[k*4+m]=(float)cdata[2*k];
#endif
          }
          for(k=0; k<nfft; k++){  // normalize
             spows[k*4+m]/=pmax;
             // if(k<10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
             // if(k>nfft-10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
          }
             
       }
         
    } else {

        /* convert power to dB  then dBm */       
        cofst=OFFSET1;       //1 bit sampling full power convert to 0 dB
        if( adbit==8){cofst=OFFSET8; }    // 8bit sampling offset
        if( adbit==4){cofst=OFFSET4; }    // 4bit sampling offset
        if( adbit==2){cofst=OFFSET2; }    // 4bit sampling offset
        for(m=0; m<numch; m++){
            for(k=0; k<nfft; k++){
                spows[k*4+m]/=scount[m]; // nomalization
                spows[k*4+m]=(float)(10.0*log10(spows[k*4+m]) + cofst); 
                if(k==0){
                    pmaxdbm=spows[k*4+m];
                    pmindbm=spows[k*4+m];
                    if(m==0){
                        ymaxdb=pmaxdbm;
                        ymindb=pmindbm;
                    }
                } else {
                    if(spows[k*4+m] > pmaxdbm) pmaxdbm=spows[k*4+m]; 
                    if(spows[k*4+m] < pmindbm) pmindbm=spows[k*4+m]; 
                }
            }
            pmax=pmaxdbm-cofst;
            pmin=pmindbm-cofst;
            printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);      
            if(pmaxdbm > ymaxdb) ymaxdb=pmaxdbm;
            if(pmindbm < ymindb) ymindb=pmindbm;
        }
            printf("All CH: Maximum and Minimum data (dBm) are %f %f\n", ymaxdb, ymindb);      

    }
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);

    /* plot spectrum (or auto cor) POSTSCRIPT OUT and XTERM*/

    for(kplot=0; kplot<2; kplot++){
       if(pmode==1 && kplot==1) goto bypass;
       if(pmode==2 && kplot==0) goto bypass;
       //if (cpgopen("?") < 1)
       if(kplot==0) {
          if (cpgopen("/PS") < 1) goto exitend;
       }
       if(kplot==1) {
          if (cpgopen(pgdevice) < 1) goto exitend;
       }

       cpgsubp(2,2);  // divide plot pannel into 2X2

       for(m=0; m<numch; m++){
           sprintf(lab,"CH#:%1d %1dbit %s sampling, Tavg(sec)=%3d\0",m+1,
                      adbit,labfsampl,sekibun);
           //printf("%s\n",lab);
           for(k=0; k<nfft; k++) {
                xdata[k]=spows[k*4+m];
           }
           if(amode > 0) {
                plotautocor(xdata, nfft, sfreq, adbit, lab, lab2, amode);
           } else {
                plotspe_a(xdata, nfft, sfreq, adbit, lab, lab2, mode2, ymaxdb,ymindb);
           }

       }
       /* panel undivided to date write */
       cpgsubp(1,1);
       /* date etc write */
       cpgiden();
       /* close the graphics device */
       cpgclos();
 bypass:;
    }

 exitend: ;
      
    free(K5.ubuf);
    free(K5.idata);
    free(cdata);
    free(spows);
    free(xdata);
    free(ydata);

#ifdef USE_FFTW
    fftw_destroy_plan(pfcc);
    fftw_destroy_plan(pbcc);
    fftw_free(cc);
#endif

 exitend2: ;

    return 0;
}
